#include <stdio.h>
#include <sys/types.h>
#include <pthread.h>
#include <service.h>
#include <socket.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <expec.h>

struct rpc_message
{
    int                 rpc_id;
    char*               data;
    int                 length;
    struct rpc_message* next;
};

static void* ext_counter_thread( void* arg );
static void announce( const char* classname );
static struct rpc_message* get_response( int rpc_id );
static void waitForEmptyReturnValue( int rpc_id );
static uint64_t waitForUint64Value( int rpc_id );
static double waitForDoubleValue( int rpc_id );
static double waitForSingleRequestResult( int rpc_id );
static char* waitForPtrValue( int rpc_id );
static time_line* waitForTimeLine( int rpc_id );
static string_list* waitForStringList( int rpc_id );

static struct rpc_message* root_message       = NULL;
static int                 is_initialized     = 0;
static int                 is_ready           = 0;
static int                 is_async           = 0;
static int                 rpc_id             = 0;
static int                 done               = 0;
static pthread_mutex_t     message_lock       = PTHREAD_MUTEX_INITIALIZER;
static expec_data_callback newDataCallback    = NULL;
static expec_init_callback initCallback       = NULL;
static expec_cl_callback   clCallback         = NULL;
static expec_sid_callback  sidCallback        = NULL;
static expec_disconnect_callback discCallback = NULL;
static pthread_t           thread;
static char               *my_server          = NULL;
static int                 thread_created     = 0;

typedef enum {MODE_COUNTER_LISTENER,MODE_COUNTER_SOURCE} op_mode;
static op_mode mode;

static void onConnect()
{
    is_ready=0;       
    if ( is_async && !thread_created )
    {
        if ( pthread_create( &thread, NULL, ext_counter_thread, NULL ) != 0 )
        {
            die( "unable to create the pthread for the remote counters" );
        }

        thread_created = 1;
    }
    if( mode==MODE_COUNTER_LISTENER )
        announce( "CounterMonitor" );
    else
        announce( "PerformanceDataSource" );
    
    if ( is_async )
    {
        while ( !is_ready )
        {
            usleep( 10 );
        }
    }
}


void expec_initialize( const char*        server,
                       int                port, 
                       expec_init_callback init_cb,
                       char               post_mortem )
{
    if ( is_initialized )
    {
        return;
    }
    is_initialized = 1;
    is_async       = ( post_mortem == 0 ) ? 1 : 0;
    root_message   = NULL;
    initCallback   = init_cb;
    mode           = MODE_COUNTER_LISTENER;
    my_server      = strdup(server);

    set_connect_notifier( onConnect );
    connectToServer( my_server, port );
}


void expec_counter_source_initialize(
		               const char*        server,
                       int                port,
                       expec_cl_callback  counter_list_cb,
                       expec_sid_callback set_id_cb)
{
    if ( is_initialized )
    {
        return;
    }
    is_initialized = 1;
    is_async       = 1;
    root_message   = NULL;
    initCallback   = NULL;
    clCallback     = counter_list_cb;
    sidCallback    = set_id_cb;
    mode           = MODE_COUNTER_SOURCE;
    my_server      = strdup(server);

    set_connect_notifier( onConnect );
    connectToServer( my_server, port );
}

void expec_finalize()
{
    done = 1;
    if ( is_async )
    {
        ( void )pthread_join( thread, NULL );
        pthread_detach( thread );
    }
    closeConnection();
    if( my_server!=NULL )
        free( my_server );
}

static void handle_incomming_clistener(const char *msg)
{
    if ( msg[ 0 ] != FUNCTION_NEW_DATA )
    {
        printf( "unhandled incomming function call with ID %d\n", msg[ 0 ] );
    	//char tmp[100];
		//snprintf( tmp, 100, "unhandled incomming function call with ID %d\n", msg[ 0 ] );
        //die( tmp );
		return;	
    }
	// unpack and call callback function
	int pos = 5;

	if ( msg[ pos ] != TYPE_UINT64_T )
	{
		die( "wrong data type (first param) in FUNCTION_NEW_DATA" );
	}
	pos += 3;
	uint64_t id = readUint64_t( msg, pos );
	pos += 8; // 3 bytes header + 8 bytes data

	if ( msg[ pos ] != TYPE_UINT64_T )
	{
		die( "wrong data type (second param) in FUNCTION_NEW_DATA" );
	}
	pos      += 3;
	uint64_t timestamp = readUint64_t( msg, pos );
	pos      += 8; // 3 bytes header + 8 bytes data

	if ( msg[ pos ] != TYPE_DOUBLE )
	{
		die( "wrong data type (third param) in FUNCTION_NEW_DATA" );
	}
	pos  += 3;
	double value = readDouble( msg, pos );

        pos  += 8; // 3 bytes header + 8 bytes data

	if ( newDataCallback != NULL )
	{
		newDataCallback( id, timestamp, value );
	}
}

static void handle_incomming_csource(const char *msg)
{
    if ( msg[ 0 ] == FUNCTION_LIST_SOURCES )
    {
    	int pos = 1;
    	uint32_t rpc_id = readUint32_t( msg, pos );
    	// nothing to extract, just read the counter list
    	// and send it as an reply
    	string_list *counter=clCallback();

    	// send reply
        int total_size = MESSAGE_ENVELOPE_SIZE;
        total_size += PARAMETER_ENVELOPE_SIZE;
        total_size += 2+string_list_sock_len( counter );

        clearOutputBuffer();
        // size
        addOutUint32( total_size );
        // call number
        addOutByte(( char )FUNCTION_RETURN );
        // unique rpc id
        addOutUint32( rpc_id );
        // return value type
        addOutByte(( char )TYPE_STRINGLIST );
        // parameter size
        addOutUint16( 2+string_list_sock_len( counter ) );
        // value itself
        addOutStringList( counter );
        // OK, and GO
        socketSendAll();
    }
    else if ( msg[ 0 ] == FUNCTION_SET_SOURCE_ID )
    {
    	// read and decode, no reply needed
		// and call callback function
		int pos = 5;

		if ( msg[ pos ] != TYPE_PTR )
		{
			die( "wrong data type (first param) in FUNCTION_SET_SOURCE_ID" );
		}
		pos += 1;
		uint32_t size = readUint16_t( msg, pos );
		pos += 2;

		const char *name = &msg[pos];

		pos += size; // size includes \0 already

		if ( msg[pos] != TYPE_UINT64_T )
		{
			die( "wrong data type (second param) in FUNCTION_SET_SOURCE_ID" );
		}
		pos      += 3;
		uint64_t id = readUint64_t( msg, pos );

		sidCallback( name, id );
    }
    else
    {
		printf( "unhandled incomming function call with ID %d\n", msg[ 0 ] );
    	//char tmp[100];
        //snprintf( tmp, 100, "unhandled incomming function call with ID %d\n", msg[ 0 ] );
        //die( tmp );
    }
}

void expec_newData( uint64_t dataSourceID,
                    uint64_t timestamp,
                    double value )
{   
    //printf("expec_newData src: %lu, ts: %lu, val: %f\n", dataSourceID, timestamp, value);
    int my_rpc_id  = rpc_id++;
    int total_size = MESSAGE_ENVELOPE_SIZE+
                     3*(PARAMETER_ENVELOPE_SIZE+8);
    clearOutputBuffer();
    // message envelope
    addOutUint32( total_size );
    addOutByte(( char )FUNCTION_CS_NEW_DOUBLE_DATA );
    addOutUint32( my_rpc_id );
    // parameter 1 envelope
    addOutByte(( char )TYPE_UINT64_T );
    addOutUint16( 8 );
    // parameter 1
    addOutUint64( dataSourceID );
    // parameter 2 envelope
    addOutByte(( char )TYPE_UINT64_T );
    addOutUint16( 8 );
    // parameter 2
    addOutUint64( timestamp );
    // parameter 3 envelope
    addOutByte(( char )TYPE_DOUBLE );
    addOutUint16( 8 );
    // parameter 3
    addOutDouble( value );
    socketSendAll();
}

static void read_and_handle_one_message()
{
    uint32_t            size;
    char*               msg;
    struct rpc_message* rpc;
    struct rpc_message* rpc_current;

    /* max wait 100 millisec for a message */
    msg = readMessage( &size, 100000 );
    if ( msg == NULL )
    {
        return;
    }
    //printf("incomming message of size %d\n", size);

    if ( msg[ 0 ] == FUNCTION_RETURN )
    {
        rpc = malloc( sizeof( struct rpc_message ));
        if ( rpc == NULL )
        {
            die( "unable to allocate memory for new message" );
        }
        rpc->data   = msg;
        rpc->length = size;
        rpc->rpc_id = readUint64_t( msg, 1 );
        rpc->next   = NULL;
        pthread_mutex_lock( &message_lock );
        if ( root_message == NULL )
        {
            root_message = rpc;
        }
        else
        {
            rpc_current = root_message;
            while ( rpc_current->next != NULL )
            {
                rpc_current = rpc_current->next;
            }
            rpc_current->next = rpc;
        }
        pthread_mutex_unlock( &message_lock );
    }
    else 
    {
        if( mode==MODE_COUNTER_LISTENER )
    	    handle_incomming_clistener(msg);
        else if( mode==MODE_COUNTER_SOURCE )
    	    handle_incomming_csource(msg);

        free( msg );
    }
}

static void* ext_counter_thread( void* arg )
{
    // call an external function at first, this can be anything
    // or an call to VTThrd_registerThread() in our case
    if ( initCallback != NULL )
    {
        initCallback( "ExtCounter" );
    }

    while ( 1 )
    {
        read_and_handle_one_message();
        /* leave if requested */
        if ( done == 1 )
        {
            break;
        }
    }

    return NULL;
}

static void announce( const char* classname )
{
    int my_rpc_id;
    int total_size;
    int str_len = strlen( classname );

    my_rpc_id   = rpc_id++;
    total_size  = MESSAGE_ENVELOPE_SIZE+PARAMETER_ENVELOPE_SIZE;
    total_size += str_len+1;

    clearOutputBuffer();
    addOutUint32( total_size );
    addOutByte(( char )FUNCTION_INIT );
    addOutUint32( my_rpc_id );
    // parameter envelope
    addOutByte(( char )TYPE_PTR );
    addOutUint16( str_len+1 );
    // parameter
    addOutString( classname );
    // send it ...
    socketSendAll();

    waitForEmptyReturnValue( my_rpc_id );

    is_ready = 1;
}

static struct rpc_message* is_response_there( int rpc_id )
{
    struct rpc_message* current_message = root_message;

    while ( current_message != NULL )
    {
        if ( current_message->rpc_id == rpc_id )
        {
            return current_message;
        }
        current_message = current_message->next;
    }
    return NULL;
}


static struct rpc_message* get_response( int rpc_id )
{
    struct rpc_message* retval = NULL;

    /** non-blocking in the async case, blocking otherwise */
    if ( is_async )
    {
        pthread_mutex_lock( &message_lock );
        retval = is_response_there( rpc_id );
        pthread_mutex_unlock( &message_lock );
    }
    else
    {
        /* wait here for message until we have it
         * no concurrent threads involved, so no
         * locking necessary
         */

        while (( retval = is_response_there( rpc_id )) == NULL )
        {
            read_and_handle_one_message();
        }
    }

    return retval;
}

static void delete_message( int rpc_id )
{
    pthread_mutex_lock( &message_lock );
    struct rpc_message* current_message = root_message;
    struct rpc_message* last_message    = NULL;

    while ( current_message != NULL )
    {
        if ( current_message->rpc_id == rpc_id )
        {
            if ( last_message == NULL )
            {
                // need to replace root
                root_message = current_message->next;
                pthread_mutex_unlock( &message_lock );
                free( current_message->data );
                free( current_message );
                return;
            }
            else
            {
                last_message->next = current_message->next;
                pthread_mutex_unlock( &message_lock );
                free( current_message->data );
                free( current_message );
                return;
            }
            break;
        }
        last_message    = current_message;
        current_message = current_message->next;
    }

    pthread_mutex_unlock( &message_lock );
}

static void waitForEmptyReturnValue( int rpc_id )
{
    struct rpc_message* result;

    while (( result = get_response( rpc_id )) == NULL )
    {
        usleep( 10 );
    }
    if ( result->length != 9 )
    {
        die( "response has wrong length, wanted 9" );
    }

    delete_message( rpc_id );
}

static uint64_t waitForUint64Value( int rpc_id )
{
    struct rpc_message* result;
    int                 pos;
    uint64_t            retval;

    while (( result = get_response( rpc_id )) == NULL )
    {
        usleep( 10 );
    }

    pos    = 5;
    retval = check_and_readUint64_t( result->data, pos );

    delete_message( rpc_id );
    return retval;
}

static double waitForDoubleValue( int rpc_id )
{
    struct rpc_message* result;
    int                 pos;
    double           retval;

    while (( result = get_response( rpc_id )) == NULL )
    {
        usleep( 10 );
    }

    pos    = 5;
    retval = check_and_readDouble( result->data, pos );

    delete_message( rpc_id );
    return retval;
}


static double waitForSingleRequestResult( int rpc_id )
{
    struct rpc_message* result;
    int                 pos;
    double           retval;

    while (( result = get_response( rpc_id )) == NULL )
    {
        usleep( 10 );
    }

    pos    = 5;
    retval = check_and_readSingleRequestResult( result->data, pos );

    delete_message( rpc_id );
    return retval;
}

static time_line* waitForTimeLineRequestResult( int rpc_id )
{
    struct rpc_message* result;
    int                 pos;
    time_line*           retval;

    while (( result = get_response( rpc_id )) == NULL )
    {
        usleep( 10 );
    }

    pos    = 5;
    retval = check_and_readTimeLineRequestResult( result->data, pos );

    delete_message( rpc_id );
    return retval;
}



static char* waitForPtrValue( int rpc_id )
{
    struct rpc_message* result;
    int                 pos;
    char*               ptr;

    while (( result = get_response( rpc_id )) == NULL )
    {
        usleep( 10 );
    }

    pos = 5;
    ptr = strdup( check_and_readPtr( result->data, pos ));

    delete_message( rpc_id );
    return ptr;
}

static time_line* waitForTimeLine( int rpc_id )
{
    struct rpc_message* result;
    int                 pos;
    time_line*          retval;

    while (( result = get_response( rpc_id )) == NULL )
    {
        usleep( 10 );
    }

    pos    = 5;
    retval = check_and_readTimeLine( result->data, pos );

    delete_message( rpc_id );
    return retval;
}

static string_list* waitForStringList( int rpc_id )
{
    struct rpc_message* result;
    int                 pos;
    string_list*        retval;

    while (( result = get_response( rpc_id )) == NULL )
    {
        usleep( 10 );
    }

    pos    = 5;
    retval = check_and_readStringList( result->data, pos );

    delete_message( rpc_id );
    return retval;
}

/**
 * sends a LIST_COUNTER_REQUEST to the server and waits for the
 * reply
 * @return a list of strings with counter names
 */
string_list* expec_listCounter()
{
    int total_size = MESSAGE_ENVELOPE_SIZE;
    int my_rpc_id  = rpc_id++;
    clearOutputBuffer();
    addOutUint32( total_size );
    addOutByte(( char )FUNCTION_LIST_COUNTER );
    addOutUint32( my_rpc_id );
    socketSendAll();
    return waitForStringList( my_rpc_id );
}

/**
 * sends the subscribe request to the server and waits
 * for the reply
 * @param counter the counter name we want to subscribe to
 * @return 0 on failure, >0 otherwise (=counter ID)
 */
uint64_t expec_subscribe( char* counter,
                         char  post_mortem )
{
    int my_rpc_id  = rpc_id++;
    int cntr_len   = strlen( counter );
    int total_size = MESSAGE_ENVELOPE_SIZE+
                     PARAMETER_ENVELOPE_SIZE+
                     cntr_len+1+
                     PARAMETER_ENVELOPE_SIZE+
                     1;
    clearOutputBuffer();
    // message envelope
    addOutUint32( total_size );
    addOutByte(( char )FUNCTION_SUBSCRIBE );
    addOutUint32( my_rpc_id );
    // parameter 1 envelope
    addOutByte(( char )TYPE_PTR );
    addOutUint16( cntr_len+1 );
    // parameter 1
    addOutString( counter );
    // parameter 2 envelope
    addOutByte(( char )TYPE_CHAR );
    addOutUint16( 1 );
    // parameter 2
    addOutByte( post_mortem );
    socketSendAll();
    return waitForUint64Value( my_rpc_id );
}

char* expec_description( char* counter )
{
    int my_rpc_id  = rpc_id++;
    int cntr_len   = strlen( counter );
    int total_size = MESSAGE_ENVELOPE_SIZE+
                     PARAMETER_ENVELOPE_SIZE+
                     cntr_len+1;
    clearOutputBuffer();
    // message envelope
    addOutUint32( total_size );
    addOutByte(( char )FUNCTION_DESCRIPTION );
    addOutUint32( my_rpc_id );
    // parameter 1 envelope
    addOutByte(( char )TYPE_PTR );
    addOutUint16( cntr_len+1 );
    // parameter 1
    addOutString( counter );
    socketSendAll();
    return waitForPtrValue( my_rpc_id );
}

char* expec_unit( char* counter )
{
    int my_rpc_id  = rpc_id++;
    int cntr_len   = strlen( counter );
    int total_size = MESSAGE_ENVELOPE_SIZE+
                     PARAMETER_ENVELOPE_SIZE+
                     cntr_len+1;
    clearOutputBuffer();
    // message envelope
    addOutUint32( total_size );
    addOutByte(( char )FUNCTION_UNIT );
    addOutUint32( my_rpc_id );
    // parameter 1 envelope
    addOutByte(( char )TYPE_PTR );
    addOutUint16( cntr_len+1 );
    // parameter 1
    addOutString( counter );
    socketSendAll();
    return waitForPtrValue( my_rpc_id );
}

double expec_storageRequestSingle(char * request)
{
	int my_rpc_id = rpc_id++;
	int reqlen = strlen(request);
	int total_size=MESSAGE_ENVELOPE_SIZE+
		PARAMETER_ENVELOPE_SIZE+
		reqlen+1+
		PARAMETER_ENVELOPE_SIZE+
		1;
	clearOutputBuffer();
	// message envelope
	addOutUint32(total_size);
	addOutByte(FUNCTION_DATA_REQUEST);
	addOutUint32(my_rpc_id);
	// parameter 1 envelope 
	addOutByte(TYPE_PTR);
	addOutUint16(reqlen+1);
	// parameter 1
	addOutString(request);
	// parameter 2 envelope 
	addOutByte(TYPE_CHAR);
	addOutUint16(1);
	// parameter 2
	addOutByte(0);   	
	socketSendAll();

	return waitForSingleRequestResult(my_rpc_id);
}

time_line* expec_storageRequestTimeLine(char * request)
{
	int my_rpc_id = rpc_id++;
	int reqlen = strlen(request);
	int total_size=MESSAGE_ENVELOPE_SIZE+
		PARAMETER_ENVELOPE_SIZE+
		reqlen+1+
		PARAMETER_ENVELOPE_SIZE+
		1;
	clearOutputBuffer();
	// message envelope
	addOutUint32(total_size);
	addOutByte(FUNCTION_DATA_REQUEST);
	addOutUint32(my_rpc_id);
	// parameter 1 envelope 
	addOutByte(TYPE_PTR);
	addOutUint16(reqlen+1);
	// parameter 1
	addOutString(request);
	// parameter 2 envelope 
	addOutByte(TYPE_CHAR);
	addOutUint16(1);
	// parameter 2
	addOutByte(0);   	
	socketSendAll();

	return waitForTimeLineRequestResult(my_rpc_id);
}


/**
 * sends the unsubscribe request to the server
 * @param counter the counter id we want to unsubscribe from
 */
time_line* expec_unsubscribe( uint64_t counter_id )
{
    int my_rpc_id  = rpc_id++;
    int total_size = MESSAGE_ENVELOPE_SIZE+
                     PARAMETER_ENVELOPE_SIZE+8;
    clearOutputBuffer();
    // message envelope
    addOutUint32( total_size );
    addOutByte(( char )FUNCTION_UNSUBSCRIBE );
    addOutUint32( my_rpc_id );
    // parameter envelope
    addOutByte(( char )TYPE_UINT64_T );
    addOutUint16( 8 );
    // parameter
    addOutUint64( counter_id );
    socketSendAll();
    return waitForTimeLine( my_rpc_id );
}

uint64_t expec_getTimestamp( uint64_t counter_id )
{
    int my_rpc_id  = rpc_id++;
    int total_size = MESSAGE_ENVELOPE_SIZE+
            PARAMETER_ENVELOPE_SIZE+8;
    clearOutputBuffer();
    // message envelope
    addOutUint32( total_size );
    addOutByte(( char )FUNCTION_GET_TIMESTAMP );
    addOutUint32( my_rpc_id );
    // parameter envelope
    addOutByte(( char )TYPE_UINT64_T );
    addOutUint16( 8 );
    // parameter
    addOutUint64( counter_id );
    socketSendAll();
    return waitForUint64Value( my_rpc_id );
}

uint64_t expec_counterStatus( char* name )
{
    int my_rpc_id  = rpc_id++;
    int total_size = MESSAGE_ENVELOPE_SIZE+
                     PARAMETER_ENVELOPE_SIZE;
    int cntr_len = strlen( name );
    total_size += cntr_len+1;
    clearOutputBuffer();
    // message envelope
    addOutUint32( total_size );
    addOutByte(( char )FUNCTION_COUNTER_STATUS );
    addOutUint32( my_rpc_id );
    // parameter envelope
    addOutByte(( char )TYPE_PTR );
    addOutUint16( cntr_len+1 );
    // parameter
    addOutString( name );
    socketSendAll();
    return waitForUint64Value( my_rpc_id );
}

void expec_registerNewDataCallback( expec_data_callback callback )
{
    newDataCallback = callback;
}

void expec_unregisterNewDataCallback( expec_data_callback callback )
{
    if ( newDataCallback != callback )
    {
        die( "will not unregister wrong callback" );
    }
    newDataCallback = NULL;
}

void expec_setDisconnetCallback( expec_disconnect_callback newcb )
{
    discCallback = newcb;
    set_disconnect_notifier( newcb );
}
